using System;
using System.IO;
using System.Text;
using org.bouncycastle.crypto;
using org.bouncycastle.crypto.parameters;
using org.bouncycastle.crypto.paddings;
using org.bouncycastle.crypto.modes;
using org.bouncycastle.util.encoders;
using org.bouncycastle.crypto.engines;
using org.bouncycastle.crypto.generators;
using org.bouncycastle.security;


namespace DESExample
{
	/**
	 * DESExample is a simple DES based encryptor/decryptor.
	 *
	 * The program is command line driven, with the input
	 * and output files specified on the command line.
	 *
	 * DESExample infile outfile [keyfile]
	 *
	 * A new key is generated for each encryption, if key is not specified,
	 * then the example will assume encryption is required, and as output
	 * create deskey.dat in the current directory.  This key is a hex
	 * encoded byte-stream that is used for the decryption.  The output
	 * file is Hex encoded, 60 characters wide text file.
	 *
	 * When encrypting;
	 *	. the infile is expected to be a byte stream (text or binary)
	 *  . there is no keyfile specified on the input line
	 *
	 * When decrypting;
	 *  . the infile is expected to be the 60 character wide base64 
	 *    encoded file
	 *  . the keyfile is expected to be a base64 encoded file
	 *
	 * This example shows how to use the light-weight API, DES and
	 * the filesystem for message encryption and decryption.
	 *
	 */
	class DESExample
	{
		// Encrypting or decrypting ?
		private bool encrypt = true;

		// To hold the initialised DESede cipher
		private PaddedBufferedBlockCipher cipher = null;

		// The input stream of bytes to be processed for encryption
		private BufferedStream inStream = null;

		// The output stream of bytes to be procssed
		private BufferedStream outStream = null;

		// The key
		private byte[] key = null;

		[STAThread]
		static void Main(string[] args)
		{
			bool   encrypt = true;
			String infile = null;
			String outfile = null;
			String keyfile = null;

			if (args.Length < 2)
			{
				Console.WriteLine("Usage: DESExample "+
					" infile outfile [keyfile]");
				Environment.Exit(1);
			}

			keyfile = "deskey.dat";
			infile = args[0];
			outfile = args[1];

			if (args.Length > 2)
			{
				encrypt = false;
				keyfile = args[2];
			}

			DESExample de = new DESExample(infile, outfile, keyfile, encrypt);
			de.process();
		}

		// Default constructor, used for the usage message
		public DESExample() { }

		/*
		 * Constructor, that takes the arguments appropriate for
		 * processing the command line directives.
		 */
		public DESExample(
			String infile,
			String outfile,
			String keyfile,
			bool encrypt
			)
		{
			/* 
			 * First, determine that infile & keyfile exist as appropriate.
			 *
			 * This will also create the BufferedStream as required
			 * for reading the input file.  All input files are treated
			 * as if they are binary, even if they contain text, it's the
			 * bytes that are encrypted.
			 */
			this.encrypt = encrypt;
			try
			{
				inStream = new BufferedStream(
					File.Open (infile, FileMode.Open, FileAccess.Read));
			}
			catch (FileNotFoundException fnf)
			{
				Console.WriteLine("Input file not found ["+infile+"]:"+
					fnf.Message);
				Environment.Exit(1);
			}

			try
			{
				outStream = new BufferedStream(
					File.Open (outfile, FileMode.Create, FileAccess.Write));
			}
			catch (IOException fnf)
			{
				Console.WriteLine("Output file not found ["+outfile+"]:"+
					fnf.Message);
				Environment.Exit(1);
			}

			if (encrypt)
			{
				try
				{
					/*
					 * The process of creating a new key requires a 
					 * number of steps.
					 *
					 * First, create the parameters for the key generator
					 * which are a secure random number generator, and
					 * the length of the key (in bits).
					 */
					SecureRandom sr = null;
					try
					{
						sr = new SecureRandom();
						/*
						 * This following call to setSeed() makes the
						 * initialisation of the SecureRandom object
						 * _very_ fast, but not secure AT ALL.  
						 *
						 * Remove the line, recreate the class file and 
						 * then run DESExample again to see the difference.
						 *
						 * The initialisation of a SecureRandom object
						 * can take 5 or more seconds depending on the
						 * CPU that the program is running on.  That can
						 * be annoying during unit testing.
						 * 	-- jon
						 */
						sr.setSeed(getBytes("www.bouncycastle.org"));
					}
					catch (Exception nsa)
					{
						Console.WriteLine("Hmmm, no SHA1PRNG, you need the "+
							"Sun implementation: "+ nsa.Message);
						Environment.Exit(1);
					}
					KeyGenerationParameters kgp = new KeyGenerationParameters(
						sr, 
						DESedeParameters.DES_EDE_KEY_LENGTH*8);

					/*
					 * Second, initialise the key generator with the parameters
					 */
					DESedeKeyGenerator kg = new DESedeKeyGenerator();
					kg.init(kgp);

					/*
					 * Third, and finally, generate the key
					 */
					key = kg.generateKey();

					/*
					 * We can now output the key to the file, but first
					 * hex encode the key so that we can have a look
					 * at it with a text editor if we so desire
					 */
					BufferedStream keystream = new BufferedStream(
						File.Open (keyfile, FileMode.Create, FileAccess.Write));
					byte[] keyhex = Hex.encode(key);
					keystream.Write(keyhex, 0, keyhex.Length);
					keystream.Flush();
					keystream.Close();
				}
				catch (IOException createKey)
				{
					Console.WriteLine("Could not decryption create key file "+
						"["+keyfile+"]: " + createKey.Message);
					Environment.Exit(1);
				}
			}
			else
			{
				try
				{
					// read the key, and decode from hex encoding
					BufferedStream keystream = new BufferedStream(
						File.Open (keyfile, FileMode.Open, FileAccess.Read));
					int len = Convert.ToInt32(keystream.Length);
					byte[] keyhex = new byte[len];
					keystream.Read(keyhex, 0, len);
					key = Hex.decode(keyhex);
				}
				catch (IOException ioe)
				{
					Console.WriteLine("Decryption key file not found, "+
						"or not valid ["+keyfile+"]"+ ioe.Message);
					Environment.Exit(1);
				}
			}
		}

		private void process()
		{
			/* 
			* Setup the DESede cipher engine, create a PaddedBufferedBlockCipher
			* in CBC mode.
			*/
			cipher = new PaddedBufferedBlockCipher(
				new CBCBlockCipher(new DESedeEngine()));

			/*
			* The input and output streams are currently set up
			* appropriately, and the key bytes are ready to be
			* used.
			*
			*/

			if (encrypt)
			{
				performEncrypt(key);
			}
			else
			{
				performDecrypt(key);
			}

			// after processing clean up the files
			try
			{
				inStream.Close();
				outStream.Flush();
				outStream.Close();
			}
			catch (IOException closing)
			{
				Console.WriteLine("Could not close files: " 
					+ closing.Message);
			}
		}
		
		/*
		 * This method performs all the encryption and writes
		 * the cipher text to the buffered output stream created
		 * previously.
		 */
		private void performEncrypt(byte[] key)
		{
			// initialise the cipher with the key bytes, for encryption
			cipher.init(true, new KeyParameter(key));

			/*
			* Create some temporary byte arrays for use in
			* encryption, make them a reasonable size so that
			* we don't spend forever reading small chunks from
			* a file.
			*
			* There is no particular reason for using getBlockSize()
			* to determine the size of the input chunk.  It just
			* was a convenient number for the example.  
			*/
			//int inBlockSize = cipher.getBlockSize() * 5;
			int inBlockSize = 47;
			int outBlockSize = cipher.getOutputSize(inBlockSize);
	
			byte[] inblock = new byte[inBlockSize];
			byte[] outblock = new byte[outBlockSize];

			/* 
			* now, read the file, and output the chunks
			*/
			try
			{
				int inL;
				int outL;
				byte[] rv = null;
				while ((inL=inStream.Read(inblock, 0, inBlockSize)) > 0)
				{
					outL = cipher.processBytes(inblock, 0, inL, outblock, 0);
					/*
					* Before we write anything out, we need to make sure
					* that we've got something to write out. 
					*/
					if (outL > 0)
					{
						rv = Hex.encode(outblock, 0, outL);
						outStream.Write(rv, 0, rv.Length);
						outStream.WriteByte(Convert.ToByte('\n'));
					}
				}

				try
				{
					/*
					* Now, process the bytes that are still buffered
					* within the cipher.
					*/
					outL = cipher.doFinal(outblock, 0);
					if (outL > 0)
					{
						rv = Hex.encode(outblock, 0, outL);
						outStream.Write(rv, 0, rv.Length);
						outStream.WriteByte(Convert.ToByte('\n'));
					}
				}
				catch (CryptoException ce)
				{

				}
			}
			catch (IOException ioeread)
			{
				Console.WriteLine(ioeread.StackTrace);
			}
		}

		/*
		 * This method performs all the decryption and writes
		 * the plain text to the buffered output stream created
		 * previously.
		 */
		private void performDecrypt(byte[] key)
		{	
			// initialise the cipher for decryption
			cipher.init(false, new KeyParameter(key));

			/* 
			* As the decryption is from our preformatted file,
			* and we know that it's a hex encoded format, then
			* we wrap the InputStream with a BufferedReader
			* so that we can read it easily.
			*/
			StreamReader br = new StreamReader(inStream);
			/* 
			* now, read the file, and output the chunks
			*/
			try
			{
				int outL;
				byte[] inblock = null;
				byte[] outblock = null;
				String rv = null;
				while ((rv = br.ReadLine()) != null)
				{
					inblock = Hex.decode(rv);
					outblock = new byte[cipher.getOutputSize(inblock.Length)];

					outL = cipher.processBytes(inblock, 0, inblock.Length, 
						outblock, 0);
					/*
					* Before we write anything out, we need to make sure
					* that we've got something to write out. 
					*/
					if (outL > 0)
					{
						outStream.Write(outblock, 0, outL);
					}
				}

				try
				{
					/*
					* Now, process the bytes that are still buffered
					* within the cipher.
					*/
					outL = cipher.doFinal(outblock, 0);
					if (outL > 0)
					{
						outStream.Write(outblock, 0, outL);
					}
				}
				catch (CryptoException ce)
				{

				}
			}
			catch (IOException ioeread)
			{
				Console.WriteLine(ioeread.StackTrace);
			}
		}
		private byte [] getBytes(String str)
		{
			char[]  cs = str.ToCharArray();
			byte[]  bs = new byte[cs.Length];

			for (int i = 0; i != cs.Length; i++)
			{
				bs[i] = Convert.ToByte(cs[i]);
			}
			return bs;
		}

		private String ByteArray2String(byte[] bs)
		{
			StringBuilder s = new StringBuilder();
			for (int i = 0; i < bs.Length; i++) 
			{
				s.Append(Convert.ToChar(bs[i]));
			}
			return s.ToString();
		}

	}
}
