using System;
using System.IO;
using System.Collections;
using org.bouncycastle.io;

namespace org.bouncycastle.asn1
{
	/**
	* a general purpose ASN.1 decoder - note: this class differs from the
	* others in that it returns null after it has read the last object in
	* the stream. If an ASN.1 NULL is encountered a DER/BER Null object is
	* returned.
	*/
	public class ASN1InputStream: DERInputStream
	{
		private class DEReos : DERObject
		{
			public override void encode(DEROutputStream outStream)
			{
				throw new IOException("Eeek!");
			}
		}

		private DERObject END_OF_STREAM = new DEReos();
		bool eofFound = false;

		public ASN1InputStream(
			Stream ins) : base(ins) {}

		protected override int readLength()
			//throws IOException
		{
			int length = read();
			if (length < 0)
			{
				throw new IOException("EOF found when length expected");
			}

			if (length == 0x80)
			{
				return -1;      // indefinite-length encoding
			}

			if (length > 127)
			{
				int size = length & 0x7f;

				length = 0;
				for (int i = 0; i < size; i++)
				{
					int next = read();

					if (next < 0)
					{
						throw new IOException("EOF found reading length");
					}

					length = (length << 8) + next;
				}
			}

			return length;
		}

		protected override void readFully(
			byte[]  bytes)
			//throws IOException
		{
			int     left = bytes.Length;
			int     len;

			if (left == 0)
			{
				return;
			}

			while ((len = read(bytes, bytes.Length - left, left)) > 0)
			{
				if ((left -= len) == 0)
				{
					return;
				}
			}

			if (left != 0)
			{
				return;
				//throw new IOException("EOF encountered in middle of object");
			}
		}

		/**
		* build an object given its tag and a byte stream to construct it
		* from.
		*/
		protected override DERObject buildObject(
			int	    tag,
			byte[]	bytes)
			//throws IOException
		{
			if (tag == DERTags.NULL)
				return new DERNull();   
			else if (tag == (DERTags.SEQUENCE | DERTags.CONSTRUCTED))
			{
				MemoryStream    bIn = new MemoryStream(bytes);
				ASN1InputStream         aIn = new ASN1InputStream(bIn);
				ASN1EncodableVector     v = new ASN1EncodableVector();

				DERObject   obj = aIn.readObject();

				while (obj != null)
				{
					v.add(obj);
					obj = aIn.readObject();
				}

				return new DERSequence(v);
			}
			else if (tag == (DERTags.SET | DERTags.CONSTRUCTED))
			{
				MemoryStream bIn = new MemoryStream(bytes);
				ASN1InputStream aIn = new ASN1InputStream(bIn);
				ASN1EncodableVector v = new ASN1EncodableVector();

				DERObject obj = aIn.readObject();

				while (obj != null)
				{
					v.add(obj);
					obj = aIn.readObject();
				}

				return new DERSet(v);
			}
			else if (tag == DERTags.BOOLEAN)
				return new DERBoolean(bytes);
			else if (tag == DERTags.INTEGER)
				return new DERInteger(bytes);
			else if (tag == DERTags.ENUMERATED)
				return new DEREnumerated(bytes);
			else if (tag == DERTags.OBJECT_IDENTIFIER)
				return new DERObjectIdentifier(bytes);
			else if (tag == DERTags.BIT_STRING)
			{
				int     padBits = bytes[0];
				byte[]  data = new byte[bytes.Length - 1];

				Array.Copy(bytes, 1, data, 0, bytes.Length - 1);

				return new DERBitString(data, padBits);
			}
			else if (tag == DERTags.UTF8_STRING)
				return new DERUTF8String(bytes);
			else if (tag == DERTags.PRINTABLE_STRING)
				return new DERPrintableString(bytes);
			else if (tag == DERTags.IA5_STRING)
				return new DERIA5String(bytes);
			else if (tag == DERTags.T61_STRING)
				return new DERT61String(bytes);
			else if (tag == DERTags.VISIBLE_STRING)
				return new DERVisibleString(bytes);
			else if (tag == DERTags.UNIVERSAL_STRING)
				return new DERUniversalString(bytes);
			else if (tag == DERTags.BMP_STRING)
				return new DERBMPString(bytes);
			else if (tag == DERTags.OCTET_STRING)
				return new DEROctetString(bytes);
			else if (tag == DERTags.UTC_TIME)
				return new DERUTCTime(bytes);
			else if (tag == DERTags.GENERALIZED_TIME)
				return new DERGeneralizedTime(bytes);
			else
			{
				//
				// with tagged object tag number is bottom 5 bits
				//
				if ((tag & DERTags.TAGGED) != 0)  
				{
					int tagNo = tag & 0x1f;

					if (tagNo == 0x1f)
					{
						int idx = 0;

						tagNo = 0;

						while ((bytes[idx] & 0x80) != 0)
						{
							tagNo |= (bytes[idx++] & 0x7f);
							tagNo <<= 7;
						}

						tagNo |= (bytes[idx] & 0x7f);

						byte[]  tmp = bytes;

						bytes = new byte[tmp.Length - (idx + 1)];
						Array.Copy(tmp, idx + 1, bytes, 0, bytes.Length);
					}

					if (bytes.Length == 0)        // empty tag!
					{
						if ((tag & DERTags.CONSTRUCTED) == 0)
						{
							return new DERTaggedObject(false, tagNo, new DERNull());
						}
						else
						{
							return new DERTaggedObject(false, tagNo, new DERSequence());
						}
					}

					//
					// simple type - implicit... return an octet string
					//
					if ((tag & DERTags.CONSTRUCTED) == 0)
					{
						return new DERTaggedObject(false, tagNo, new DEROctetString(bytes));
					}

					MemoryStream bIn = new MemoryStream(bytes);
					ASN1InputStream aIn = new ASN1InputStream(bIn);

					DEREncodable dObj = aIn.readObject();

					//
					// explicitly tagged (probably!) - if it isn't we'd have to
					// tell from the context
					//
					if (aIn.available() == 0)
					{
						return new DERTaggedObject(tagNo, dObj);
					}

					//
					// another implicit object, we'll create a sequence...
					//
					ASN1EncodableVector v = new ASN1EncodableVector();

					while (dObj != null)
					{
						v.add(dObj);
						dObj = aIn.readObject();
					}

					return new DERTaggedObject(false, tagNo, new DERSequence(v));
				}

				return new DERUnknownTag(tag, bytes);
			}
		}

		/**
		* read a string of bytes representing an indefinite length object.
		*/
		private byte[] readIndefiniteLengthFully()
			//throws IOException
		{
			MemoryStream   bOut = new MemoryStream();
			int                     b, b1;

			b1 = read();

			while ((b = read()) >= 0)
			{
				if (b1 == 0 && b == 0)
				{
					break;
				}

				bOut.WriteByte((byte)b1);
				b1 = b;
			}

			return bOut.ToArray();
		}

		private BERConstructedOctetString buildConstructedOctetString()
			//throws IOException
		{
			ArrayList                  octs = new ArrayList();

			for (;;)
			{
				DERObject		o = readObject();

				if (o == END_OF_STREAM)
				{
					break;
				}

				octs.Add(o);
			}

			return new BERConstructedOctetString(octs);
		}

		public override DERObject readObject()
			//throws IOException
		{
			int tag = read();
			if (tag == -1)
			{
				if (eofFound)
				{
					throw new EOFException("attempt to read past end of file.");
				}

				eofFound = true;

				return null;
			}
	    
			int     length = readLength();

			if (length < 0)    // indefinite length method
			{
				if (tag == DERTags.NULL)
					return new BERNull();
				else if (tag == (DERTags.SEQUENCE | DERTags.CONSTRUCTED))
				{
					ASN1EncodableVector  v = new ASN1EncodableVector();
	    
					for (;;)
					{
						DERObject   obj = readObject();

						if (obj == END_OF_STREAM)
						{
							break;
						}

						v.add(obj);
					}
					return new BERSequence(v);
				}
				else if (tag == (DERTags.SET | DERTags.CONSTRUCTED))
				{
					ASN1EncodableVector v = new ASN1EncodableVector();
	    
					for (;;)
					{
						DERObject   obj = readObject();

						if (obj == END_OF_STREAM)
						{
							break;
						}

						v.add(obj);
					}
					return new BERSet(v);
				}
				else if (tag == (DERTags.OCTET_STRING | DERTags.CONSTRUCTED))
					return buildConstructedOctetString();
				else
				{
					//
					// with tagged object tag number is bottom 5 bits
					//
					if ((tag & DERTags.TAGGED) != 0)  
					{
						int tagNo = tag & 0x1f;

						if (tagNo == 0x1f)
						{
							int b = read();

							tagNo = 0;

							while ((b >= 0) && ((b & 0x80) != 0))
							{
								tagNo |= (b & 0x7f);
								tagNo <<= 7;
								b = read();
							}

							tagNo |= (b & 0x7f);
						}

						//
						// simple type - implicit... return an octet string
						//
						if ((tag & DERTags.CONSTRUCTED) == 0)
						{
							byte[]  bytes = readIndefiniteLengthFully();

							return new BERTaggedObject(false, tagNo, new DEROctetString(bytes));
						}

						//
						// either constructed or explicitly tagged
						//
						DERObject		dObj = readObject();

						if (dObj == END_OF_STREAM)     // empty tag!
						{
							return new DERTaggedObject(tagNo);
						}

						DERObject       next = readObject();

						//
						// explicitly tagged (probably!) - if it isn't we'd have to
						// tell from the context
						//
						if (next == END_OF_STREAM)
						{
							return new BERTaggedObject(tagNo, dObj);
						}

						//
						// another implicit object, we'll create a sequence...
						//
						ASN1EncodableVector v = new ASN1EncodableVector();

						v.add(dObj);

						do
						{
							v.add(next);
							next = readObject();
						}
						while (next != END_OF_STREAM);

						return new BERTaggedObject(false, tagNo, new BERSequence(v));
					}

					throw new IOException("unknown BER object encountered");
				}
			}
			else
			{
				if (tag == 0 && length == 0)    // end of contents marker.
				{
					return END_OF_STREAM;
				}

				byte[]  bytes = new byte[length];
	    
				readFully(bytes);
	    
				return buildObject(tag, bytes);
			}
		}
	}

}