using System;
using System.IO;
using System.Collections;
//using org.bouncycastle.io;

namespace org.bouncycastle.asn1
{
	/**
	* Generalized time object.
	*/
	public class DERGeneralizedTime: DERObject
	{
		String      time;

		/**
		* return a generalized time from the passed in object
		*
		* @exception ArgumentException if the object cannot be converted.
		*/
		public static DERGeneralizedTime getInstance(
			Object  obj)
		{
			if ((obj == null) || (typeof(DERGeneralizedTime).IsInstanceOfType(obj)))
			{
				return (DERGeneralizedTime)obj;
			}

			if (typeof(ASN1OctetString).IsInstanceOfType(obj))
			{
				return new DERGeneralizedTime(((ASN1OctetString)obj).getOctets());
			}

			throw new ArgumentException("illegal object in getInstance: " + obj.GetType().ToString());
		}

		/**
		* return a Generalized Time object from a tagged object.
		*
		* @param obj the tagged object holding the object we want
		* @param explicit true if the object is meant to be explicitly
		*              tagged false otherwise.
		* @exception ArgumentException if the tagged object cannot
		*               be converted.
		*/
		public static DERGeneralizedTime getInstance(
			ASN1TaggedObject obj,
			bool          flagExplicit)
		{
			return getInstance(obj.getObject());
		}

		/**
		* The correct format for this is YYYYMMDDHHMMSSZ, or without the Z
		* for local time, or Z+-HHMM on the end, for difference between local
		* time and UTC time.
		* <p>
		*
		* @param time the time string.
		*/
		public DERGeneralizedTime(
			String  time)
		{
			this.time = time;
		}

		internal DERGeneralizedTime(
			byte[]  bytes)
		{
			//
			// explicitly convert to characters
			//
			char[]  dateC = new char[bytes.Length];

			for (int i = 0; i != dateC.Length; i++)
			{
				dateC[i] = (char)(bytes[i] & 0xff);
			}

			this.time = new String(dateC);
		}

		/**
		* return the time - always in the form of 
		*  YYYYMMDDhhmmssGMT(+hh:mm|-hh:mm).
		* <p>
		* Normally in a certificate we would expect "Z" rather than "GMT",
		* however adding the "GMT" means we can just use:
		* <pre>
		*     dateF = new SimpleDateFormat("yyyyMMddHHmmssz");
		* </pre>
		* To read in the time and get a date which is compatible with our local
		* time zone.
		*/
		public String getTime()
		{
			//
			// standardise the format.
			//
			if (time.Length == 15)
			{
				return time.Substring(0, 14) + "GMT+00:00";
			}
			else if (time.Length == 17)
			{
				return time.Substring(0, 14) + "GMT" + time.Substring(15, 2) + ":" + time.Substring(17, 2);
			}

			return time;
		}

		private byte[] getOctets()
		{
			char[]  cs = time.ToCharArray();
			byte[]  bs = new byte[cs.Length];

			for (int i = 0; i != cs.Length; i++)
			{
				bs[i] = (byte)cs[i];
			}

			return bs;
		}


		public override void encode(DEROutputStream  outStream)
			//throws IOException
		{
			outStream.writeEncoded(DERTags.GENERALIZED_TIME, this.getOctets());
		}
	}

}