using System;
using System.IO;
//using System.Collections;
using org.bouncycastle.io;
using org.bouncycastle.asn1;

namespace org.bouncycastle.asn1.x509
{
	public class AlgorithmIdentifier : DEREncodable
	{
		private DERObjectIdentifier objectId;
		private DEREncodable        parameters;
		private bool             parametersDefined = false;
		
		public static AlgorithmIdentifier getInstance(
			ASN1TaggedObject obj,
			bool          flagExplicit)
		{
			return getInstance(ASN1Sequence.getInstance(obj, flagExplicit));
		}
	    
		public static AlgorithmIdentifier getInstance(
			Object  obj)
		{
			if (typeof(AlgorithmIdentifier).IsInstanceOfType(obj))
			{
				return (AlgorithmIdentifier)obj;
			}

			if (typeof(DERObjectIdentifier).IsInstanceOfType(obj))
			{
				return new AlgorithmIdentifier((DERObjectIdentifier)obj);
			}

			if (typeof(String).IsInstanceOfType(obj))
			{
				return new AlgorithmIdentifier((String)obj);
			}

			if (typeof(ASN1Sequence).IsInstanceOfType(obj))
			{
				return new AlgorithmIdentifier((ASN1Sequence)obj);
			}

			throw new ArgumentException("unknown object in factory");
		}

		public AlgorithmIdentifier(
			DERObjectIdentifier     objectId)
		{
			this.objectId = objectId;
		}

		public AlgorithmIdentifier(
			String     objectId)
		{
			this.objectId = new DERObjectIdentifier(objectId);
		}

		public AlgorithmIdentifier(
			DERObjectIdentifier     objectId,
			DEREncodable            parameters)
		{
			parametersDefined = true;
			this.objectId = objectId;
			this.parameters = parameters;
		}

		public AlgorithmIdentifier(
			ASN1Sequence   seq)
		{
			objectId = (DERObjectIdentifier)seq.getObjectAt(0);

			if (seq.size() == 2)
			{
				parametersDefined = true;
				parameters = seq.getObjectAt(1);
			}
			else
			{
				parameters = null;
			}
		}

		public virtual DERObjectIdentifier getObjectId()
		{
			return objectId;
		}

		public virtual DEREncodable getParameters()
		{
			return parameters;
		}

		/**
		* <pre>
		*      AlgorithmIdentifier ::= SEQUENCE {
		*                            algorithm OBJECT IDENTIFIER,
		*                            parameters ANY DEFINED BY algorithm OPTIONAL }
		* </pre>
		*/
		public virtual DERObject getDERObject()
		{
			ASN1EncodableVector  v = new ASN1EncodableVector();

			v.add(objectId);

			if (parametersDefined)
			{
				v.add(parameters);
			}

			return new DERSequence(v);
		}

		public override bool Equals(
			Object  o)
		{
			
			if ((o == null) || !(typeof(AlgorithmIdentifier).IsInstanceOfType(o)))
			{
				return false;
			}

			AlgorithmIdentifier other = (AlgorithmIdentifier)o;

			if (!this.getObjectId().Equals(other.getObjectId()))
			{
				return false;
			}

			if (this.getParameters() == null && other.getParameters() == null)
			{
				return true;
			}

			if (this.getParameters() == null || other.getParameters() == null)
			{
				return false;
			}

			MemoryStream   b1Out = new MemoryStream();
			MemoryStream   b2Out = new MemoryStream();
			DEROutputStream         d1Out = new DEROutputStream(b1Out);
			DEROutputStream         d2Out = new DEROutputStream(b2Out);

			try
			{
				d1Out.writeObject(this.getParameters());
				d2Out.writeObject(other.getParameters());

				byte[]  b1 = b1Out.ToArray();
				byte[]  b2 = b2Out.ToArray();

				if (b1.Length != b2.Length)
				{
					return false;
				}

				for (int i = 0; i != b1.Length; i++)
				{
					if (b1[i] != b2[i])
					{
						return false;
					}
				}
			}
			catch (Exception e)
			{
				e.ToString();
				return false;
			}

			return true;
		}
		public override int GetHashCode()
		{
			return base.GetHashCode();
		}

	}

}